/*  =========================================================================================
 *   
 *    CS408 Environmental Monitoring Independent of Existing Infrastructure
 *    Copyright (C) 2021 Callum Inglis
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License along
 *    with this program; if not, write to the Free Software Foundation, Inc.,
 *    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *    
 *    Contact: Callum.Inglis.2018(at)uni.strath.ac.uk
 *    
 *  =========================================================================================
 *    
 *  Libraries Used
 *    ArduinoJson v6 Library - https://arduinojson.org/v6/doc/ (https://github.com/sandeepmistry/arduino-LoRa/blob/master/LICENSE)
 *    LoRa Library - https://github.com/sandeepmistry/arduino-LoRa/blob/master/API.md (https://github.com/bblanchon/ArduinoJson/blob/6.x/LICENSE.md)
 *    SHTSensor - https://github.com/Sensirion/arduino-sht (https://github.com/Sensirion/arduino-sht/blob/master/LICENSE)
 *
 *  =========================================================================================
 */

#include <SPI.h>
#include <LoRa.h>
#include <string>
#include <ESP8266WiFi.h>
#include ".\secrets.h"

#define ss 5
#define rst 0
#define dio0 4

// WiFi Config
char ssid[] = SECRET_SSID;
char pass[] = SECRET_PASS;
char host[] = SECRET_HOSTNAME; 
int wifiTimeout = 60000; // ms
bool wifiEstablished = false;

// LoRa Message Tracking
byte localAddress = 0xFF;
int recipient;
String incoming;
byte sender;
byte incomingMsgId;
byte incomingLength;
byte incomingMessage;

// Sensor Values
byte temperature;
byte humidity;
byte co2;
byte ppm10;
byte ppm25; //TODO
byte ppm100; // TODO

bool setupLoRa() {
  LoRa.setPins(ss, rst, dio0);  
  if (!LoRa.begin(433E6)) {
    Serial.println("[-] Fatal. Starting LoRa failed!");
    return false;
  }
  
  Serial.println("[+] LoRa Initialized OK!");
  return true;
}

/**
 * Attempt to establish WiFi connection, 
 *  account for timeout handler.
 *  
 * @return bool Connection Succesful
 */
bool setupWifi() {
  Serial.print("Connecting to ");
  Serial.print(ssid);

  // Connect
  WiFi.hostname(host);
  WiFi.begin(ssid, pass);

  // Await Connection
  while (WiFi.status() != WL_CONNECTED) {
    if (wifiTimeout <= 0) {
      Serial.println();
      Serial.print("Unable to connect to WiFi");
      return false;
    }
    
    delay(500); // Wait and retry
    wifiTimeout -= 500;
    Serial.print(".");
  }
  
  Serial.println();
  Serial.println("WiFi connected");
  Serial.print("IP address: ");
  Serial.print(WiFi.localIP());
  Serial.println();

  return true;
}

bool publishLogToAPI(char* sensorID, double temperature, double humidity, double co2, double ppm10, double ppm25, double ppm100) {
  if (!wifiEstablished) {
    Serial.println("[x] Wifi Connection Required");
    return false;
  }
  
  WiFiClientSecure client;
  client.setInsecure();
  Serial.print("[+] Connecting To ");
  Serial.println(SECRET_API_HOST);

  if (!client.connect(SECRET_API_HOST, SECRET_API_PORT)) {
    Serial.println("[x] Connection to API failed");
    return false;
  }

  // TODO ppm10, ppm100
  // Build API URL
  String url = "/api/sensor/logValue/";
  url.concat(sensorID);
  url.concat("/");
  url.concat(temperature);
  url.concat("/");
  url.concat(humidity);
  url.concat("/");
  url.concat(co2);
  url.concat("/");
  url.concat(ppm25);

  Serial.print("[+] Requesting API URL: ");
  Serial.print(SECRET_API_HOST);
  Serial.print(url);
  Serial.print(" on port ");
  Serial.println(SECRET_API_PORT);

  // Send Request TODO Change to POST
  Serial.println("[+] Sending GET request...");
  client.print(String("GET ") + url + " HTTP/1.1\r\n" +
               "Host: " + SECRET_API_HOST + "\r\n" +
               "User-Agent: ESP8266" + SECRET_HOSTNAME + "\r\n" +
               "Connection: close\r\n\r\n");
  Serial.println("[+] Request sent");

  // Check response header for 200 status Code
  String line = client.readStringUntil('\n');
  if (line.startsWith("HTTP\/1\.1 200 OK")) {
    Serial.println("[+] Upload Successful!");
    
  } else {
    Serial.println("[!] Error During Upload");
    client.stop();
    return false;
  }

  // Skip Rest of Header
  while (client.connected()) {
    line = client.readStringUntil('\n');
    if (line == "\r") {
      break;
    }
  }

  // Print out reply for debugging
  line = client.readStringUntil('\n');
  Serial.println("[+] Reply was:");
  Serial.println("==========");
  Serial.println(line);
  Serial.println("==========");
  Serial.println("[+] Closing connection");
  client.stop();

  return true;
}

void setup() {
  delay(1000);
  
  Serial.begin(115200);
  Serial.println("[+] Receiver & API Gateway");

  // Setup LoRa
  if (!setupLoRa()) { while(1); } // Die on error

  // Setup Wifi
  wifiEstablished = setupWifi();
}

void loop() {
  //publishLogToAPI("1", 2, 3, 4, 5, 6, 7);
  //delay(5000);

  // TODO Re-introduce LoRa receiver, call publishLogToAPI() with response data
  int packetSize = LoRa.parsePacket();
  if (packetSize) {
//    
//    // received a packet
    Serial.println("********************************************");
    Serial.print("[+] Received packet '");
//
//    // read packet
    //Serial.print("[+] Packet: ");
    //Serial.println(LoRa.read());

    //Serial.println("Avg PPM 1.0: " + LoRa.read());
    //Serial.println("Avg PPM 2.5: " + LoRa.read());
    //Serial.println("Avg PPM 10.0: " + LoRa.read());
    //Serial.println("Avg Temp: " + LoRa.read());
    //Serial.println("Avg Humidity: " + LoRa.read());
//
    String incoming = "";
//    std::string incoming = "";

//    String incoming = "";
    while (LoRa.available()) {
      incoming.concat((char)LoRa.read());
//      std::string newVal = (char*) LoRa.read();

//      incoming = incoming+newVal;
    }
    Serial.print(incoming);
    //Serial.print(incoming.c_str());

    // Deal with Message, delimiter is comma
//    char* msg;
//    msg = strtok(incoming, ",");
//    while (msg != NULL) {
//      printf("%s\n", msg);
//      msg = strtok(NULL, ",");
//    }
//    
    
//
//    // print RSSI of packet
    Serial.print("' with RSSI ");
    Serial.println(LoRa.packetRssi());
    Serial.println("");
  }

  
  //int packetSize = LoRa.parsePacket();
  //Serial.println(packetSize);
  //if (packetSize == 0) return;
  
  //recipient = LoRa.read();
  //sender = LoRa.read();
  //incomingMsgId = LoRa.read();
  //incomingLength = LoRa.read();
  //incomingMessage = LoRa.read();
  
  //incoming = String(incomingMessage);
  
  //if (incomingLength != incoming.length()) {
  //  Serial.println("[-] Message length does not match length");
  //  return;
  //}
  
  //if (recipient != localAddress) {
  //  Serial.println("[i] This message is not for me.");
  //  return;
  //}
  
  //Serial.println("********************************************");
  //Serial.println("Received from: 0x" + String(sender, HEX));
  //Serial.println("Sent to: 0x" + String(recipient, HEX));
  //Serial.println("Message ID: " + String(incomingMsgId));
  //Serial.println("Message length: " + String(incomingLength));
  //Serial.println("Message: " + String(incomingMessage));
  //Serial.println("RSSI: " + String(LoRa.packetRssi()));
  //Serial.println("Snr: " + String(LoRa.packetSnr()));
  //Serial.println("");
}
